/********************************************************
 * linux/kernel/bossa.c 
 *
 *   Copyright (C) 2001 INRIA, EMN, DIKU. All rights reserved.
 *   Institut National de Recherche en Informatique et en Automatique.
 *   Ecole des Mines de Nantes
 *   University of Copenhagen
 *
 *  This file should be included by the Bossa scheduler module
 * 
 *  Bossa run time system - events, event notification, functions
 *
 ********************************************************/

#define EXPORTS xxx // for bossa include in sched.h

#include <linux/init.h> /* For __init */
#include <linux/errno.h> /* Miscellaneous error codes */
#include <linux/stddef.h> /* NULL */
#include <linux/slab.h> /* kmalloc() */
#include <linux/module.h> /* EXPORT_SYMBOL */
#include <linux/smp_lock.h> /* for spinlock */
#include <linux/time.h>   /* for timer functions */
#include <linux/timer.h>  /* for timer functions */
#ifdef CONFIG_BOSSA_HRT
#include <linux/hrtime.h> /* for timer functions */
#endif

// for interface functions
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/unistd.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/uaccess.h>
#include <linux/ioctl.h>

// for /proc interface
#include <linux/proc_fs.h>

#define true 1
#define false 0

#ifdef CONFIG_BOSSA_LIBRARY
#define PROTOCOL_BOSSA_STRUCT bossa_struct
#include "library.h"
#endif


#define attr(x) list_entry(x, struct task_struct, bossa)

//#include <linux/bossa.h> /* Bossa definitions, included from sched.h */

unsigned long flags;

extern struct bossa_struct *running;

#define event_mask0(e) ((e) & 0xff000000)
#define event_mask1(e) ((e) & 0xffff0000)
#define event_mask2(e) ((e) & 0xffffff00)
#define event_mask3(e) ((e) & 0xffffffff)

spinlock_t ebuffer_lock = SPIN_LOCK_UNLOCKED;  /* second */

spinlock_t bossa_scheduler_lock = SPIN_LOCK_UNLOCKED;  /* lock for the 
bossa scheduler */

struct bossa_struct  *rts_idle_process;     // System wide idle
enum SchedState root_s_state;              // state of the root scheduler
struct bossa_struct *running;

//extern int __init bossa_Linux_ioctl_init(void);

/*******************************************************
 *
 * Bossa Event Buffer declaration and related variables
 *
 *     The event buffer is a circular buffer implemented as an array 
 *     of fixed size (bounded by MAX_EVENT_BUFFER_SIZE)
 *     It is statically allocated in the kernel
 *
 *******************************************************/

// careful: event buffer must be static
//static struct event_struct bossa_event_buffer[MAX_EVENT_BUFFER_SIZE];


struct trace_struct 
{
  int type;
  int pid;
};

// buffer to trace bossa events (CONFIG_BOSSA_TRACE)

//static struct trace_struct t_buffer[3000];

volatile int  readpos;             /* read position, first unread event 
in the buffer */
volatile int  writepos;            /* write position, first free entry 
in the buffer */
volatile long bossa_event_counter; /* global event counter */
volatile long buffer_entries;      /* number of event entries in the 
event buffer */

int bossa_trace_events = 0;  /* flag to turn on event tracing - for 
debugging */
int bossa_trace_event_counter = 0;

/*
 *  Bossa root scheduler
 */

struct scheduler_struct root_scheduler;
int root_scheduler_index;

extern struct scheduler_operations linux_sched_ops;
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDFLinux
extern struct scheduler_operations edflinux_sched_ops;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDFRTLinux
extern struct scheduler_operations edfrtlinux_sched_ops;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Progress
extern struct scheduler_operations progress_sched_ops;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Timer
extern struct scheduler_operations timer_sched_ops;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDF
extern struct scheduler_operations edf_sched_ops;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Fixed_priority
extern struct scheduler_operations fixed_priority_sched_ops;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Proportion
extern struct scheduler_operations proportion_sched_ops;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Proportion_prio
extern struct scheduler_operations proportion_prio_sched_ops;
#endif
#ifdef CONFIG_BOSSA_LIBRARY
extern struct scheduler_operations linux_library_sched_ops;
#endif

#define MAX_SCHEDULERS 10
#define MAX_SCHED_NAME_LENGTH 40
#define MAX_SCHED_PATH_LENGTH 256
int bossa_scheduler_requested = 0;
char *bossa_scheduler_name;
extern int bossa_scheduler_requested;
extern char *bossa_scheduler_name;
static int default_scheduler;
static int schedulers = 0;
struct scheduler_info {
  int root;
  int loaded;
  struct scheduler_operations *ops;
  struct proc_dir_entry* father_proc_entry; // /proc dir of parent
  struct proc_dir_entry* sched_proc_entry;  // /proc dir of self
};
static struct scheduler_info all_schedulers[MAX_SCHEDULERS];

static int __init bossa_manager_ioctl_init(void);

/********************************************************************************
*	Declaration Function to build the /proc/bossa subsystem			*
*****************************************************************************/
static struct proc_dir_entry *proc_bossa = NULL; // the root

static void make_static_proc(void);
static void bossa_create_proc_entry(struct scheduler_info *sched,
				    struct proc_dir_entry *dir);
static void find_child_entry(struct scheduler_info *sched);
static void bossa_remove_proc_entry(struct scheduler_info *sched,
				    int was_loaded, int was_root);
static void bossa_create_tree_entry(struct scheduler_info *sched_child);
static void bossa_change_proc_root(struct scheduler_info *sched);
static struct scheduler_info *find_sched_info(struct scheduler_operations *);
static void find_path_symlink(struct scheduler_info *sched_child,
			      char *path, char *name);

/****************************************************************************/


/*******************************************************
 *
 * bossa_notify_event()
 * 
 *     stores an event in the event buffer
 *
 * Parameters :
 *    event_type, target process, source process
 * Return values :
 *     -EOVERFLOW : buffer overflow
 *******************************************************/

struct next_list_struct no_next_list;

/*
For the source process, the next scheduler list begins with the current
scheduler.
For the target process, the next scheduler list begins with the child of the
current scheduler.
This is because for the source, we only want to know if it is managed by
the current scheduler, and for the target we only want to know how to get
to the destination of forwardImmediate.
*/

int bossa_notify_event (int event_type, struct bossa_struct * 
			target, struct bossa_struct * source) 
{
  unsigned long flags;
  struct event_struct e;
  struct next_list_struct *x;

  spin_lock_irqsave(&bossa_scheduler_lock,flags);

  e.type =  event_type;
  //e.time = jiffies;
  e.source = source;
  e.target = target;
  //  e.counter = bossa_event_counter;

  if (source == NULL)
    x = &no_next_list;
  else {
    x = e.source->next_list->next;
  }

#ifdef CONFIG_BOSSA_LIBRARY
#ifdef CONFIG_BOSSA_LIBRARY_CHECKS
  switch (event_mask0(event_type)) {
  case EVENT_PROCESS: {
    unsigned int event_42 = event_mask1(event_type);
    if (event_42 == EVENT_PROCESS_END)
      pre_check_process_end(target);
    else if (event_42 == EVENT_PROCESS_NEW)
      pre_check_process_new(source,target);
    else panic("unknown event %x\n",event);
    break;
  }
  case EVENT_SYSTEM:
    pre_check_system_clocktick(target);
    break;
  case EVENT_UNBLOCK:
    pre_check_unblock_nonpreemptive(target);
    break;
  case EVENT_YIELD:
    pre_check_yield(target);
    break;
  default: panic("unknown event %x\n",event);
  }
#endif
#endif

  {
  root_s_state =
    root_scheduler.ops->handle_event(&e,x,e.target->next_list->next->next);
  }

#ifdef CONFIG_BOSSA_LIBRARY
#ifdef CONFIG_BOSSA_LIBRARY_CHECKS
  post_check();
#endif
  root_s_state = compute_scheduler_state();
#endif

  spin_unlock_irqrestore(&bossa_scheduler_lock,flags);

  return 0;
}

int rts_ready(void) { return root_s_state == READY; }
int rts_blocked(void) { return root_s_state == BLOCKED; }

/*******************************************************
 *
 * bossa_check_process_state()
 * 
 *     check whether the state of a process is in
 *     one of a set of classes, represented as an or'd
 *     mask.
 *
 * Parameters :
 *    target process, classes
 * Return values :
 *     0, non-zero
 *******************************************************/

int bossa_check_process_state (struct bossa_struct * target,
			       unsigned int classes) 
{
  int res;
  unsigned long flags;


  spin_lock_irqsave(&bossa_scheduler_lock,flags);
  res = root_scheduler.ops->check_process_state(target,
						target->next_list->next->next,
						classes);
  spin_unlock_irqrestore(&bossa_scheduler_lock,flags);

  return res;
}

static
struct bossa_struct * bossa_create_process(struct task_struct *p,
					   struct bossa_struct *parent)
{
  struct bossa_struct *new;
  struct scheduler_operations *parent_ops;

  if (parent == rts_idle_process)
    parent_ops = NULL;
  else parent_ops = parent->next_list->prev->sched->ops;

  new = (struct bossa_struct *)&(p->bossa);

   // initialize Bossa specific attributes
  if (parent_ops && parent_ops->is_default) {
#ifndef CONFIG_BOSSA_LIBRARY
    new->next_list = &(parent_ops->next_list);
#else
    library_create_process(new,&(parent_ops->next_list));
#endif
  }
  else {
#ifndef CONFIG_BOSSA_LIBRARY
    new->next_list = &(all_schedulers[default_scheduler].ops->next_list);
#else
    library_create_process
      (new,&(all_schedulers[default_scheduler].ops->next_list));
#endif
  }
  return new;
}

void rts_create_process(int caller, struct task_struct *p) {
  // replaces wake_up code, switch state before root
  p->state = TASK_RUNNING;
  bossa_create_process(p, &(current->bossa));
  if (p->pid == 1) {
    /* first process to run, attach it */
    bossa_notify_event(EVENT_PROCESS_NEW, &(p->bossa), NULL);
    if (root_s_state == READY) current->need_resched = 1;
  }
  else {
    bossa_notify_event(EVENT_PROCESS_NEW, &(p->bossa), &(current->bossa));
  }
}

void rts_terminate_process(struct task_struct *p) {
  bossa_notify_event(EVENT_PROCESS_END_USER, &(p->bossa), NULL);
}

void rts_clocktick(int caller) {
  unsigned long flags;
  spin_lock_irqsave(&bossa_scheduler_lock,flags);
  if (/*user_mode(regs) &&*/ (current->pid) && root_s_state == RUNNING) {
    bossa_notify_event(EVENT_SYSTEM_CLOCKTICK, &(current->bossa), NULL);
  }
  spin_unlock_irqrestore(&bossa_scheduler_lock,flags);
}

void rts_yield(struct task_struct *p) {
  bossa_notify_event(EVENT_YIELD_USER, &(p->bossa), NULL);
}

void rts_set_priority (struct task_struct * tgt, int niceval) 
{
  unsigned long flags;
  struct bossa_struct * target = &(tgt->bossa);

  spin_lock_irqsave(&bossa_scheduler_lock,flags);
  root_scheduler.ops->set_priority(target,target->next_list->next->next,
				   niceval);
  spin_unlock_irqrestore(&bossa_scheduler_lock,flags);
}

struct ctl_table;
int rts_scheduler_info( struct ctl_table * table, int write, struct file * filp, void * buffer, size_t * lenp )
{
	if (!*lenp || (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}
	if (!write)
        {
		int len;
		char b[4096];

		len = 0;
		memset(b,0,sizeof(b));
		len = root_scheduler.ops->do_schedprocinfo(b,1);
		if (len > *lenp)
			len = *lenp;
		if (len)
			if(copy_to_user(buffer, b, len))
				return -EFAULT;
		if (len < *lenp) {
			if(put_user('\n', ((char *) buffer) + len))
				return -EFAULT;
			len++;
		}
		*lenp = len;
		filp->f_pos += len;
	}
	else
		*lenp=0;

        return 0;
}

int rts_proc_info(struct task_struct *task, char * buffer) {
  struct next_list_struct *n = task->bossa.next_list->next;
  return (n->sched->ops->do_taskprocinfo(n,&(task->bossa),buffer,1,1,1));
}

#ifdef CONFIG_BOSSA_CENTRALIZED
#include "bossa/sched.c"
#endif



/*******************************************************
 *
 * Managing the next list
 *
 *     These functions are used in hierarchy management.
 *
 *******************************************************/


static inline void next_add_to_head_queue(struct scheduler_struct *s,
					  struct next_list_struct *q)
{
  struct next_list_struct *p;
  struct next_list_struct *next = q->next;

  p =
    (struct next_list_struct *)
    (kmalloc(sizeof(struct next_list_struct), GFP_ATOMIC));

  p->sched = s;
  p->prev = q;
  q->next = p;
  p->next = next;
  next->prev = p;
}

static inline void next_rem_from_queue(struct next_list_struct *q)
{
  struct next_list_struct *p, *n;
  p = q->prev;
  n = q->next;
  p->next = n;
  n->prev = p;
  kfree(q);
}

static struct next_list_struct init_next_list_entry;

void print_next_list(struct next_list_struct *n) {
  int i=0;
  struct next_list_struct *p;
  for(p=n->next; p!=n && i < 10; p=p->next, i++) {
    if (!p->sched)
      printk("no sched ");
    else if (!p->sched->ops)
      printk("no ops ");
    else printk("%s ",p->sched->ops->name);
  }
  printk("\n");
}

static inline void init_next_add_to_head_queue(struct scheduler_struct *s,
					  struct next_list_struct *q)
{
  struct next_list_struct *p;
  struct next_list_struct *next = q->next;

  p = &init_next_list_entry;

  p->sched = s;
  p->prev = q;
  q->next = p;
  p->next = next;
  next->prev = p;
}

// copy q into p and add a new entry for s at the end
static inline void next_add_to_tail_queue(struct next_list_struct *p,
					  struct next_list_struct *q,
					  struct scheduler_struct *s)
{
  struct next_list_struct *p1 = p;
  struct next_list_struct *r;

  for (r=q->next; r!=q; r=r->next) {
    next_add_to_head_queue(r->sched,p1);
    p1=p1->next;
  }
  next_add_to_head_queue(s,p1);
}


/*******************************************************
 * bossa_add_scheduler()
 *
 *    Initialize variables for the Run time system (event buffer) and 
 *    plugs the root scheduler
 *    - called at system startup by init/main.c via bossa_init()
 *
 * Parameters :
 *    bossa_ops
 * Return values :
 *    NONE
 *******************************************************/

#define CHECK(s1,s2) if (strcmp(s1,s2) == 0) return true;

static inline int same_name_as_a_static_schedule(char *name) {
  CHECK(name,linux_sched_ops.name);
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDFLinux
  CHECK(name,edflinux_sched_ops.name);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDFRTLinux
  CHECK(name,edfrtlinux_sched_ops.name);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Progress
  CHECK(name,progress_sched_ops.name);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Timer
  CHECK(name,timer_sched_ops.name);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDF
  CHECK(name,edf_sched_ops.name);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Fixed_priority
  CHECK(name,fixed_priority_sched_ops.name);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Proportion
  CHECK(name,proportion_sched_ops.name);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Proportion_prio
  CHECK(name,proportion_prio_sched_ops.name);
#endif
#ifdef CONFIG_BOSSA_LIBRARY
  CHECK(name,linux_library_sched_ops.name);
#endif
  return false;
}

int bossa_add_scheduler(struct scheduler_operations* ops, int stat) {
  if (!stat && same_name_as_a_static_schedule(ops->name)) {
    printk("cannot replace a static scheduler\n");
    return -1;
  }
  {
    int i = schedulers++;
    all_schedulers[i].ops = ops;
    all_schedulers[i].root = 0;
    all_schedulers[i].loaded = 0;
    all_schedulers[i].ops->init();
    
    all_schedulers[i].ops->next_list.sched = NULL;
    all_schedulers[i].ops->next_list.next =
      all_schedulers[i].ops->next_list.prev =
      &(all_schedulers[i].ops->next_list);
    
    all_schedulers[i].father_proc_entry = NULL;
    all_schedulers[i].sched_proc_entry = NULL;
    
    // create an info file for the scheduler in the root directory
    // why would proc_bossa still be NULL?
    if (proc_bossa != NULL)
      bossa_create_proc_entry(&all_schedulers[i], proc_bossa);
    return 0;
  }
}

/*******************************************************
 * bossa_remove_scheduler()
 *
 *     
 *    - remove the scheduler into all_schedulers[]
 *    - update the scheduler list
 *    
 *
 * Parameters :
 *    bossa_ops
 * Return values :
 *    NONE
 *******************************************************/
void bossa_remove_scheduler(struct scheduler_operations *u_ops) {
  int i,position = -1;
  for (i=0; i!=schedulers; i++) {
    if (all_schedulers[i].ops == u_ops) {
      position = i;
    }
  }
  if (position < 0) {
    printk("can't unloaded scheduler because it is not loaded");
    return;
  }

  bossa_remove_proc_entry(&all_schedulers[position],
			  all_schedulers[position].loaded,
			  all_schedulers[position].root);

  all_schedulers[position] = all_schedulers[schedulers-1];
  schedulers--;
}

struct scheduler_operations *find_scheduler_by_id(int position) {
  if (position >= 0 && position < schedulers &&
      all_schedulers[position].loaded && !all_schedulers[position].root) {
    return all_schedulers[position].ops;
  }
  printk("invalid scheduler %d\n",position);
  return NULL;
}

/*******************************************************
 * bossa_init()
 *
 *    Initialize variables for the Run time system (event buffer) and 
 *    plugs the root scheduler
 *    - called at system startup by init/main.c
 *
 * Parameters :
 *    NONE
 * Return values :
 *    NONE
 *******************************************************/

void  __init bossa_init(int caller)
{
  printk(KERN_EMERG "Initializing Bossa\n");

  /* plug the root scheduler */

  printk("Initializing Bossa root scheduler: \n");

  rts_idle_process = &(init_task.bossa);

  // Init running
  running = rts_idle_process;
  root_s_state = BLOCKED;

/*
  Record the set of loaded schedulers.
  There can be at most MAX_SCHEDULERS schedulers.
  Initialize the queues and process variables for each loaded scheduler.
  Initialize the root scheduler variable.
*/

  default_scheduler = -1;

#ifndef CONFIG_BOSSA_LIBRARY
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDFLinux
  bossa_add_scheduler(&edflinux_sched_ops,true);
  if (bossa_scheduler_requested && default_scheduler == -1 &&
      !strcmp(bossa_scheduler_name,edflinux_sched_ops.name))
    default_scheduler = schedulers-1;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDFRTLinux
  bossa_add_scheduler(&edfrtlinux_sched_ops,true);
  if (bossa_scheduler_requested && default_scheduler == -1 &&
      !strcmp(bossa_scheduler_name,edfrtlinux_sched_ops.name))
    default_scheduler = schedulers-1;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Progress
  bossa_add_scheduler(&progress_sched_ops,true);
  if (bossa_scheduler_requested && default_scheduler == -1 &&
      !strcmp(bossa_scheduler_name,progress_sched_ops.name))
    default_scheduler = schedulers-1;
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Timer
  bossa_add_scheduler(&timer_sched_ops,true);
  if (bossa_scheduler_requested && default_scheduler == -1 &&
      !strcmp(bossa_scheduler_name,timer_sched_ops.name))
    default_scheduler = schedulers-1;
#endif
  // use Linux if nothing else/valid is requested
  bossa_add_scheduler(&linux_sched_ops,true);
  if (/*bossa_scheduler_requested && */default_scheduler == -1
      /*&& !strcmp(bossa_scheduler_name,"linux")*/)
    default_scheduler = schedulers-1;
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_EDF
  bossa_add_scheduler(&edf_sched_ops,true);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Fixed_priority
  bossa_add_scheduler(&fixed_priority_sched_ops,true);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Proportion
  bossa_add_scheduler(&proportion_sched_ops,true);
#endif
#ifdef CONFIG_BOSSA_SCHEDULER_MODULE_Proportion_prio
  bossa_add_scheduler(&proportion_prio_sched_ops,true);
#endif
#else
    bossa_add_scheduler(&linux_library_sched_ops,true);
    default_scheduler = 0;
#endif

  all_schedulers[default_scheduler].root = true;
  all_schedulers[default_scheduler].loaded = true;
  root_scheduler_index = default_scheduler;

  root_scheduler.ops = all_schedulers[default_scheduler].ops;

  /* initialize the dummy next_list object */
  no_next_list.next = no_next_list.prev = &no_next_list;
  no_next_list.sched = NULL;
  init_next_add_to_head_queue(&root_scheduler,
			      &(all_schedulers[default_scheduler].ops->next_list));
  root_scheduler.ops->attach_self();
}

void  __init bossa_ioctl_init(int caller) {
  // initialize ioctls of all loaded schedulers
  int i;
  for (i=0; i!=schedulers; i++)
    all_schedulers[i].ops->ioctl_init();
  bossa_manager_ioctl_init();
}

// count all processes that are ready or sleeping in the Linux sense, and
// that are not nowhere or terminated in the Bossa sense
int count_all_processes(void) {
  struct task_struct *p;
  int state;
  int i=0;
  static const char * stat_nam[] = { "R ", "S ", "D ", "Z ", "T ", "W " };
  for_each_task(p) {
    state = p->state ? ffz(~p->state) + 1 : 0;
	printk("%d ", p->pid);
	state = p->state ? ffz(~p->state) + 1 : 0;
	if (((unsigned) state) < sizeof(stat_nam)/sizeof(char *))
		printk(stat_nam[state]);
	else
		printk(" ");
    
    if (state < 3 && p->bossa.state && p->bossa.state < 32) i++;
  }
  return i;
}

/* ---------------------------------------------------------------------
 * new timer library
 *
 * create_bossa_timer: used by the compiler for process timers
 * create_static_bossa_timer: used by the compiler for scheduler timers
 * create_persistent_bossa_timer: used by the compiler for persistent timers
 * delete_bossa_timer: used by the compiler for process timers
 * delete_static_bossa_timer: used by the compiler for scheduler timers
 * start_absolute_timer: used by the programmer
 * start_relative_timer: used by the programmer (add jiffies to provided time)
 * stop_timer: stops a timer
 * --------------------------------------------------------------------- */

static void new_process_bossa_timeout(unsigned long __data);
static void new_process_bossa_static_timeout(unsigned long __data);
static void new_process_bossa_persistent_timeout(unsigned long __data);

/* timer data, defined in bossa.h
struct bossa_timer_data {
  struct bossa_struct *p; // target process, or NULL if global to the policy
  struct next_list_struct *next_list; // next list leading to the policy
  unsigned int timer_index; // indicates which of the policy's timers this is
};

struct bossa_persistent_timer_data {
  struct bossa_struct *p; // target process, or NULL if global to the policy
  struct next_list_struct *next_list; // next list leading to the policy
  unsigned int timer_index; // indicates which of the policy's timers this is
  struct timer_list *timer;
  int target_alive;
};
*/

#ifdef BOSSA_HRT
// borrowed from itimer.c
int bossa_tstojiffie(unsigned long sec, long nsec_arg, unsigned long *jiff)
{
  int res = CONFIG_HIGH_RES_RESOLUTION;
	long nsec = nsec_arg + res - 1;

//	printk("sec %d nsec_arg %d nsec %d\n",sec,nsec_arg,nsec);

        /*
         * A note on jiffy overflow: It is possible for the system to
         * have been up long enough for the jiffies quanity to overflow.
         * In order for correct timer evaluations we require that the
         * specified time be somewhere between now and now + (max
         * unsigned int/2).  Times beyond this will be truncated back to
         * this value.  This is done in the absolute adjustment code,
         * below.  Here it is enough to just discard the high order
         * bits.  
         */
        *jiff = HZ * sec;
        /*
         * Do the res thing. (Don't forget the add in the declaration of nsec) 
         */
        nsec -= nsec % res;
        /*
         * Split to jiffie and sub jiffie
         */
        *jiff += nsec / (NSEC_PER_SEC / HZ);
        /*
         * We trust that the optimizer will use the remainder from the 
         * above div in the following operation as long as they are close. 
         */
/*  	printk("res %d %d\n",*jiff, */
/*  	       (nsec_to_arch_cycles(nsec % (NSEC_PER_SEC / HZ)))); */
        return  (nsec_to_arch_cycles(nsec % (NSEC_PER_SEC / HZ)));
}
#endif

struct timer_list *
create_bossa_timer(struct bossa_struct *p, struct next_list_struct *next_list,
		   unsigned int index) {
  struct timer_list *timer =
    (struct timer_list *)kmalloc(sizeof(struct timer_list),GFP_KERNEL);
  struct bossa_timer_data *timer_data =
    (struct bossa_timer_data *)kmalloc(sizeof(struct bossa_timer_data),
				       GFP_KERNEL);

  init_timer(timer);
  timer->data = (unsigned long)timer_data;
  timer_data->p = p;
  timer_data->next_list = next_list;
  timer_data->timer_index = index;
  timer->function = new_process_bossa_timeout;
  return timer;
}

struct timer_list *
create_static_bossa_timer(struct bossa_struct *p,
			  struct next_list_struct *next_list,
			  struct timer_list *timer,
			  struct bossa_timer_data *timer_data,
			  unsigned int index) {
  init_timer(timer);
  timer->data = (unsigned long)timer_data;
  timer_data->p = p;
  timer_data->next_list = next_list;
  timer_data->timer_index = index;
  timer->function = new_process_bossa_static_timeout;
  return timer;
}

struct timer_list *
create_persistent_bossa_timer(struct bossa_struct *p,
			      struct next_list_struct *next_list,
			      unsigned int index,
			      int data_sz) {
  struct timer_list *timer =
    (struct timer_list *)kmalloc(sizeof(struct timer_list),GFP_KERNEL);
  struct bossa_persistent_timer_data *timer_data =
    (struct bossa_persistent_timer_data *)
    kmalloc(sizeof(struct bossa_persistent_timer_data)+data_sz,GFP_KERNEL);

  init_timer(timer);
  timer->data = (unsigned long)timer_data;
  timer_data->p = p;
  timer_data->next_list = next_list;
  timer_data->timer_index = index;
  timer_data->timer = timer;
  timer_data->target_alive = 1/*TRUE*/;
  timer->function = new_process_bossa_persistent_timeout;
  return timer;
}

static inline int running_timer(struct timer_list *t) {
  return (t->list.next && t->list.next->prev == &(t->list));
}

void globalize_bossa_timer(struct timer_list *timer) {
  struct bossa_persistent_timer_data *timer_data =
    (struct bossa_persistent_timer_data *)(timer->data);
  timer_data->target_alive = 0/*FALSE*/;
}

/* for persistent timers also */
void delete_bossa_timer(struct timer_list *t) {
  if (running_timer(t)) {
    del_timer_sync(t);
  }
  kfree((struct bossa_timer_data *)t->data);
  kfree(t);
}

void delete_static_bossa_timer(struct timer_list *t) {
  if (running_timer(t)) {
    del_timer_sync(t);
  }
}

#ifdef CONFIG_BOSSA_HRT
void start_absolute_timer(struct timer_list *t, struct timespec *abs) {
  unsigned long abs_jiffies = abs->tv_sec;
  long abs_sub_jiffies = abs->tv_nsec;

  if (running_timer(t)) {
    mod_timer_hr(t, abs_jiffies, abs_sub_jiffies);
  }
  else {
    t->expires = abs_jiffies;
    t->sub_expires = abs_sub_jiffies;
    add_timer(t);
  }
}

void start_relative_timer(struct timer_list *t, struct timespec *rel) {
  unsigned long rel_jiffies = rel->tv_sec;
  long rel_sub_jiffies = rel->tv_nsec;
  unsigned long abs_jiffies;
  long abs_sub_jiffies;

  // The use of locks here follows the model of do_timer_gettime in itimer.c
  write_lock(&xtime_lock);
  update_jiffies_sub();
  abs_jiffies = rel_jiffies + jiffies;
  abs_sub_jiffies = rel_sub_jiffies + sub_jiffie();
  write_unlock(&xtime_lock);

  if ((unsigned long)abs_sub_jiffies > cycles_per_jiffies) {
    abs_sub_jiffies -= cycles_per_jiffies;
    abs_jiffies++;
  }

  if (running_timer(t)) {
    mod_timer_hr(t, abs_jiffies, abs_sub_jiffies);
  }
  else {
    t->expires = abs_jiffies;
    t->sub_expires = abs_sub_jiffies;
    add_timer(t);
  }
}
#else
void start_absolute_timer(struct timer_list *t, unsigned long abs_time) {
  if (running_timer(t)) {
    mod_timer(t, abs_time);
  }
  else {
    t->expires = abs_time;
    add_timer(t);
  }
}

void start_relative_timer(struct timer_list *t, unsigned long rel_time) {
  if (running_timer(t)) {
    mod_timer(t, rel_time + jiffies);
  }
  else {
    t->expires = rel_time + jiffies;
    add_timer(t);
  }
}
#endif

void stop_timer(struct timer_list *t) {
  del_timer_sync(t);
}

static void new_process_bossa_timeout(unsigned long __data) {
  unsigned long flags;
  struct bossa_timer_data * p = (struct bossa_timer_data *) __data;
  spin_lock_irqsave(&bossa_scheduler_lock,flags);

/* The following only works if the kerlogd and syslogd are turned off */
/* printk("ending timer at %d %d\n",jiffies,sub_jiffie()); */

#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_unblock_timer(p->p);
#endif
  root_s_state =
    root_scheduler.ops->unblock_timer_target_entry(p->timer_index,
						   p->p,p->next_list);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif

  spin_unlock_irqrestore(&bossa_scheduler_lock,flags);
}

static void new_process_bossa_static_timeout(unsigned long __data) {
  unsigned long flags;
  struct bossa_timer_data * p = (struct bossa_timer_data *) __data;
  spin_lock_irqsave(&bossa_scheduler_lock,flags);

/* The following only works if the kerlogd and syslogd are turned off */
/* printk("ending timer at %d %d\n",jiffies,sub_jiffie()); */

#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_unblock_timer(p->p);
#endif
  root_s_state =
    root_scheduler.ops->unblock_timer_notarget_entry(p->timer_index,
						     p->p,p->next_list);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif

  spin_unlock_irqrestore(&bossa_scheduler_lock,flags);
}

static void new_process_bossa_persistent_timeout(unsigned long __data) {
  unsigned long flags;
  struct bossa_persistent_timer_data * p =
    (struct bossa_persistent_timer_data *) __data;
  spin_lock_irqsave(&bossa_scheduler_lock,flags);

/* The following only works if the kerlogd and syslogd are turned off */
/* printk("ending timer at %d %d\n",jiffies,sub_jiffie()); */

#ifdef CONFIG_BOSSA_LIBRARY
  pre_check_unblock_timer(p->p);
#endif
  root_s_state =
    root_scheduler.ops->unblock_timer_persistent_entry(p->timer_index,
						       p->timer,
						       (((char *)p)+
							sizeof(struct bossa_persistent_timer_data)),
						       p->target_alive,
						       p->p,p->next_list);
#ifdef CONFIG_BOSSA_LIBRARY
  post_check();
  root_s_state = compute_scheduler_state();
#endif

  spin_unlock_irqrestore(&bossa_scheduler_lock,flags);
}


/* ---------------------------------------------------------------------
 * timers for Bossa policies
 * --------------------------------------------------------------------- */

void process_bossa_timeout(unsigned long __data);

struct timer_list *no_timer(void) {
  return NULL;
}

struct timer_list *init_bossa_timer(struct bossa_struct *p,
				    unsigned long timeout) {
  struct timer_list *timer;

  timer = (struct timer_list *)kmalloc(sizeof(struct timer_list),GFP_KERNEL);
  init_timer(timer);
  timer->expires = timeout + jiffies;
  timer->data = (unsigned long)p;
  timer->function = process_bossa_timeout;
  add_timer(timer);
  return timer;
}

void reset_bossa_timer(struct timer_list *timer, unsigned long timeout) {
  mod_timer(timer, timeout + jiffies);
  current->need_resched = 1;
}

void end_bossa_timer(struct timer_list *timer) {
  if (timer) {
    del_timer_sync(timer);
    kfree(timer);
  }
}

#ifdef CONFIG_BOSSA_HRT
// always immediately copied by the policy
static struct timespec now_res;

struct timespec *now(void) {
/*
  write_lock(&xtime_lock);
  update_jiffies_sub();
*/
  now_res.tv_sec = jiffies;
  now_res.tv_nsec = sub_jiffie();
/*
  write_unlock(&xtime_lock);
*/
  return &now_res;
}
#else
unsigned long now(void) {
  return jiffies;
}
#endif

void process_bossa_timeout(unsigned long __data) {
  struct bossa_struct * p = (struct bossa_struct *) __data;
  bossa_notify_event(EVENT_UNBLOCK_TIMER, p, NULL);
}


/* ---------------------------------------------------------------------
 * copy from user
 * --------------------------------------------------------------------- */

int get_user_int(void *user_int) {
  int progress;
  copy_from_user(&progress,user_int,sizeof(int));
  return progress;
}

/* ---------------------------------------------------------------------
 * interface for building a hierarchy
 * --------------------------------------------------------------------- */


static int major;
static devfs_handle_t devfs_handle;
extern spinlock_t bossa_scheduler_lock;


#define BOSSA_ID_NUM 0xCE
#define IOCTL_CONNECT        _IOWR(BOSSA_ID_NUM, 0, int)
#define IOCTL_PRINT          _IOR(BOSSA_ID_NUM, 1, int)
#define IOCTL_ATTACH_INFO    _IOWR(BOSSA_ID_NUM, 2, int)
#define IOCTL_SCHED	     _IOWR(BOSSA_ID_NUM, 3, int)
#define IOCTL_NBSCHED        _IOWR(BOSSA_ID_NUM, 4, int)
#define IOCTL_NAMES          _IOR(BOSSA_ID_NUM, 5, int)
#define IOCTL_DISCONNECT     _IOWR(BOSSA_ID_NUM, 6, int)
#define IOCTL_CONNECT_ROOT   _IOWR(BOSSA_ID_NUM, 7, int)
#define IOCTL_JINFO          _IOWR(BOSSA_ID_NUM, 8, int)
#define BOSSA_DEVICE_NAME "bossa"
#define BOSSA_DEVICE_NAME_TWO "bossa2"


#define SUCCESS 0
#define FAILURE -1
#define SIZE_ALL_SCHEDULE_NAME 30

struct dev_bossa_attach_info_struct {
  int parent;
  int parent_attach_argc;
  char parent_attach_argv[300];
};

struct dev_bossa_connect_struct {
  int parent;
  int child;
  int data[10];
};

struct dev_bossa_connect_root_struct {
  int new_root;
  int data[10];
};

struct dev_bossa_disconnect_struct {
  int tgt;
};

struct dev_bossa_print_struct {
  char *s;
  int sz;
};

struct dev_bossa_move_struct {
  int pid;
  char dest[32];
};

struct dev_bossa_jinfo_struct {
  int jinfo_cycles1;
  int jinfo_cycles2;
  int jinfo_cycles3;
  int jinfo_cycles4;
  int jinfo_cycles5;
  int jinfo_jiffies1;
  int jinfo_pages1;
  int jinfo_pages2;
  int jinfo_pages3;
  int jinfo_dpages1;
  int jinfo_dpages2;
  int jinfo_dpages3;
};

extern int jinfo_cycles1;
extern int jinfo_cycles2;
extern int jinfo_cycles3;
extern int jinfo_cycles4;
extern int jinfo_cycles5;
extern int jinfo_jiffies1;
extern int jinfo_pages1;
extern int jinfo_pages2;
extern int jinfo_pages3;
extern int jinfo_dpages1;
extern int jinfo_dpages2;
extern int jinfo_dpages3;

struct scheduler_ioctl_info {
  int root;
  int num;
  int loaded;
  enum sched_type type;
  char name[SIZE_ALL_SCHEDULE_NAME];
};

static inline struct bossa_struct *find_process_by_pid(pid_t pid) {
  struct task_struct *tsk;
  
  if (pid) {
    //read_lock_irq(&tasklist_lock);
    tsk = find_task_by_pid(pid);
    //read_unlock_irq(&tasklist_lock);
  }
  else tsk = current;
  return (struct bossa_struct *)(&tsk->bossa);
}

static inline int _dev_bossa_ioctl(struct inode *inode,
				   struct file *fd,
				   unsigned int ioctl_num,
				   unsigned long param) {
  switch (ioctl_num) {
  case IOCTL_CONNECT: {
    struct dev_bossa_connect_struct data;
    struct scheduler_info *parent, *child;
    struct scheduler_struct *child_scheduler;
    copy_from_user(&data,
		   (struct dev_bossa_connect_struct *)param,
		   sizeof(struct dev_bossa_connect_struct));
    if (data.parent >= schedulers || data.child >= schedulers ||
	data.parent < 0 || data.child < 0) {
      // proposed parent or child out of range
      printk("proposed parent or child is not a valid scheduler\n");
      errno = -EINVAL;
      return FAILURE;
    }
    parent = &all_schedulers[data.parent];
    child = &all_schedulers[data.child];
    if (parent->ops->type != VS) {
      // parent must be a VS
      printk("parent must be a VS\n");
      errno = -EINVAL;
      return FAILURE;
    }
    if (parent->loaded) { // child is new
      if (child->loaded) {
	// parent and child are both already in the hierarchy
	printk("both already loaded\n");
	errno = -EINVAL;
	return FAILURE;
      }
      if (child->ops->is_default && !parent->ops->is_default) {
	printk("default child must have a default parent\n");
	errno = -EINVAL;
	return FAILURE;
      }
      // child is new
      child_scheduler = parent->ops->create_scheduler(child->ops,0);
      child->loaded = true;
      next_add_to_tail_queue(&(child->ops->next_list),
			     &(parent->ops->next_list),
			     child_scheduler);
      child->ops->attach_self();
      /* add a /proc entry for the child below its parent */
      bossa_create_proc_entry(child, parent->sched_proc_entry);
    }
    else {
      if (!child->loaded) {
	// neither parent nor child is already in the hierarchy
	printk("no one is loaded %d %d\n",data.parent,data.child);
	errno = -EINVAL;
	return FAILURE;
      }
      // parent is new
      if (!child->root) {
	// can only add a new parent to the root
	printk("can only add a new parent to the root\n");
	errno = -EINVAL;
	return FAILURE;
      }
      if (!parent->ops->is_default) {
	printk("root scheduler must be a default scheduler\n");
	errno = -EINVAL;
	return FAILURE;
      }
      if (!child->ops->is_default) {
	printk("the only child of a default VS must be a default scheduler\n");
	errno = -EINVAL;
	return FAILURE;
      }
      child_scheduler = parent->ops->create_scheduler(child->ops,1);
      root_scheduler.ops = parent->ops;
      {
	int i;
	for (i=0; i!=schedulers; i++) {
	  // add old child after root scheduler, which has
	  // been updated to the new root
	  if (all_schedulers[i].loaded) {
	    next_add_to_head_queue(child_scheduler,
				   all_schedulers[i].ops->next_list.next);
	  }
	}
      }
      next_add_to_head_queue(&root_scheduler,
			     &(parent->ops->next_list));
      parent->ops->attach_self();
      parent->loaded = true;
      parent->root = true;
      child->root = false;
      // add a new root
      bossa_change_proc_root(&all_schedulers[root_scheduler_index]);
      root_scheduler_index = data.parent;
    }
    // assumed not to fail...
    parent->ops->attach(child->ops->next_list.prev->sched,data.data);
    return SUCCESS;
  }
  case IOCTL_CONNECT_ROOT: {
    struct dev_bossa_connect_root_struct data;
    struct scheduler_info *new_root;
    struct scheduler_info *old_root = &all_schedulers[root_scheduler_index];
    copy_from_user(&data,
		   (struct dev_bossa_connect_root_struct *)param,
		   sizeof(struct dev_bossa_connect_root_struct));
    if (data.new_root >= schedulers || data.new_root < 0) {
      // proposed root or child out of range
      printk("proposed root is not a valid scheduler\n");
      errno = -EINVAL;
      return FAILURE;
    }
    new_root = &all_schedulers[data.new_root];
    if (new_root->ops->type != VS) {
      // parent must be a VS
      printk("new root must be a VS\n");
      errno = -EINVAL;
      return FAILURE;
    }
    if (new_root->loaded) {
      // new root cannot already be in the hierarchy
      printk("new root cannot already be in the hierarchy\n");
      errno = -EINVAL;
      return FAILURE;
    }
    else {
      struct scheduler_struct *old_root_sched;
      // parent is new
      if (!new_root->ops->is_default) {
	printk("root scheduler must be a default scheduler\n");
	errno = -EINVAL;
	return FAILURE;
      }
      old_root_sched = new_root->ops->create_scheduler(root_scheduler.ops,1);
      root_scheduler.ops = new_root->ops;
      {
	int i;
	for (i=0; i!=schedulers; i++) {
	  // add old child after root scheduler, which has
	  // been updated to the new root
	  if (all_schedulers[i].loaded) {
	    next_add_to_head_queue(old_root_sched,
				   all_schedulers[i].ops->next_list.next);
	  }
	}
      }
      next_add_to_head_queue(&root_scheduler,&(new_root->ops->next_list));
      new_root->ops->attach_self();
      new_root->loaded = true;
      new_root->root = true;
      old_root->root = false;
      bossa_change_proc_root(&all_schedulers[root_scheduler_index]);
      root_scheduler_index = data.new_root;
    }
    // assumed not to fail...
    new_root->ops->attach(old_root->ops->next_list.prev->sched,data.data);
    return SUCCESS;
  }
  case IOCTL_ATTACH_INFO: {
    struct dev_bossa_attach_info_struct data;
    struct scheduler_info *parent;
    copy_from_user(&data,
		   (struct dev_bossa_attach_info_struct *)param,
		   sizeof(struct dev_bossa_attach_info_struct));
    if (data.parent >= schedulers) {
      // proposed parent or child out of range
      printk("proposed parent or child is not a valid scheduler\n");
      errno = -EINVAL;
      return FAILURE;
    }
    parent = &all_schedulers[data.parent];
    if (parent->ops->type != VS) {
      // parent must be a VS
      printk("parent must be a VS\n");
      errno = -EINVAL;
      return FAILURE;
    }
    {
      char *x[10];
      int argc = parent->ops->attach_info(x);
      int i;
      char *p = data.parent_attach_argv;
      data.parent_attach_argc = argc;
      for(i=0; i!=argc; i++,p+=30) {
	strncpy(p,x[i],29);
	p[29] = 0;
      }
      copy_to_user((struct dev_bossa_attach_info_struct *)param,
		   &data,sizeof(struct dev_bossa_attach_info_struct));
    }
    return SUCCESS;
  }
  case IOCTL_DISCONNECT: {
    struct dev_bossa_disconnect_struct data;
    struct scheduler_info *tgt;
    copy_from_user(&data,
		   (struct dev_bossa_disconnect_struct *)param,
		   sizeof(struct dev_bossa_disconnect_struct));
    if (data.tgt >= schedulers) {
      // proposed parent or child out of range
      printk("proposed target is not a valid scheduler\n");
      errno = -EINVAL;
      return FAILURE;
    }
    if (data.tgt == default_scheduler) {
      printk("cannot disconnect the default scheduler\n");
      errno = -EINVAL;
      return FAILURE;
    }
    tgt = &all_schedulers[data.tgt];
    if (tgt->loaded) {
      struct scheduler_struct *child_scheduler;
      child_scheduler = tgt->ops->detach_self(tgt->root);
      if ((int)child_scheduler == -1) {
	// errno already set
	printk("can't detach the child from its previous scheduler\n");
	return FAILURE;
      }
      tgt->loaded = false;
      if (tgt->root) {
	int i;
	tgt->root = false;
	for(i=0; i!=schedulers; i++) {
	  if (all_schedulers[i].loaded) {
	    // we don't remove the root, which points to root_scheduler
            // which will get updated to the child scheduler.
            // instead we delete the node containing the child scheduler,
            // which we have a pointer to anyway
	    next_rem_from_queue(all_schedulers[i].ops->next_list.next->next);
	  }
	  if (all_schedulers[i].ops == child_scheduler->ops) {
	    all_schedulers[i].root = true;
	    root_scheduler_index = i;
	    root_scheduler.ops = all_schedulers[i].ops;
	  }
	}
	bossa_change_proc_root(tgt);
      }
      else {
	bossa_remove_proc_entry(tgt,true,false);
      }
      bossa_create_proc_entry(tgt, proc_bossa);
      {
	struct next_list_struct *elem, *tmp_elem;
	for(elem = tgt->ops->next_list.next,
	      tmp_elem = elem->next; elem != &tgt->ops->next_list;
	    elem = elem->next, tmp_elem = tmp_elem->next) {
	  kfree(elem);
	}
	tgt->ops->next_list.next = tgt->ops->next_list.prev =
	  &(tgt->ops->next_list);
      }
    }
    else {
      printk("scheduler not in the hierarchy\n");
      errno = -EINVAL;
      return FAILURE;
    }
    return SUCCESS;
  }
  case IOCTL_PRINT: {
    struct dev_bossa_print_struct data;
    struct next_list_struct *s;
    char buffer[500];
    char *b;
    int i;
    copy_from_user(&data,
		   (struct dev_bossa_print_struct *)param,
		   sizeof(struct dev_bossa_print_struct));
    b = buffer;
    for(i=0; i!=schedulers; i++)
      b += sprintf(b,"%d. %s (%s, %s, %s)\n",i,all_schedulers[i].ops->name,
		   (all_schedulers[i].ops->type == PS ? "PS" : "VS"),
		   (all_schedulers[i].root ? "root" :
		    all_schedulers[i].loaded ? "mounted" : "not mounted"),
		   (all_schedulers[i].ops->is_default ?
		    "default" : "not default"));
    b += sprintf(b,"\nDefault path:\n");
    for(s=all_schedulers[default_scheduler].ops->next_list.next;
	s->sched->ops->type != PS;
	s = s->next) {
      b += sprintf(b,"%s -> ",s->sched->ops->name);
    }
    sprintf(b,"%s\n",s->sched->ops->name);
    copy_to_user(data.s,buffer,b-buffer > data.sz ? b-buffer : data.sz);
    return SUCCESS;
  }
  case IOCTL_SCHED: {
    struct scheduler_ioctl_info data;

    copy_from_user(&data,
		   (struct scheduler_ioctl_info *)param,
		   sizeof(struct scheduler_ioctl_info));
    
    strncpy(data.name,all_schedulers[data.num].ops->name,SIZE_ALL_SCHEDULE_NAME);
    data.type = all_schedulers[data.num].ops->type;
    data.root = all_schedulers[data.num].root;
    data.loaded = all_schedulers[data.num].loaded;
    
    copy_to_user((struct scheduler_ioctl_info *)param,
		  &data,sizeof(struct scheduler_ioctl_info));
    return SUCCESS;
  }
  case IOCTL_NBSCHED: {
    int scheduler;

    copy_from_user(&scheduler,
		   (int *)param,
		   sizeof(scheduler));
    
    scheduler = schedulers;
    
    copy_to_user((int *)param,
		   &scheduler,
		   sizeof(scheduler));
    return SUCCESS;
  }
  case IOCTL_NAMES: {
    struct dev_bossa_print_struct data;
    char buffer[500];
    char *b;
    int i;
    copy_from_user(&data,
		   (struct dev_bossa_print_struct *)param,
		   sizeof(struct dev_bossa_print_struct));
    b = buffer;
    for(i=0; i!=schedulers; i++) {
      strncpy(b,all_schedulers[i].ops->name,24);
      *(b+24) = '\0';
      b += 25;
    }
    copy_to_user(data.s,buffer,data.sz);
    return SUCCESS;
  }
  case IOCTL_JINFO: {
#ifdef CONFIG_BOSSA_TRACE
    struct dev_bossa_jinfo_struct data;
    copy_from_user(&data,
		   (struct dev_bossa_jinfo_struct *)param,
		   sizeof(struct dev_bossa_jinfo_struct));
    jinfo_cycles1 = data.jinfo_cycles1;
    jinfo_cycles2 = data.jinfo_cycles2;
    jinfo_cycles3 = data.jinfo_cycles3;
    jinfo_cycles4 = data.jinfo_cycles4;
    jinfo_cycles5 = data.jinfo_cycles5;
    jinfo_jiffies1 = data.jinfo_jiffies1;
    jinfo_pages1 = data.jinfo_pages1;
    jinfo_pages2 = data.jinfo_pages2;
    jinfo_pages3 = data.jinfo_pages3;
    jinfo_dpages1 = data.jinfo_dpages1;
    jinfo_dpages2 = data.jinfo_dpages2;
    jinfo_dpages3 = data.jinfo_dpages3;
    return SUCCESS;
#else
    printk("tracing not enabled\n");
    return FAILURE;
#endif
  }
  }
  printk("unknown ioctl\n");
  errno = -EINVAL;
  return FAILURE;
}

static int dev_bossa_ioctl(struct inode *inode,
			   struct file *fd,
			   unsigned int ioctl_num,
			   unsigned long param) {
  int res;
  unsigned long flags;
  
  spin_lock_irqsave(&bossa_scheduler_lock,flags);
  res = _dev_bossa_ioctl(inode,fd,ioctl_num,param);
  spin_unlock_irqrestore(&bossa_scheduler_lock,flags);
  return res;
}

static struct file_operations procfops = {
  ioctl: dev_bossa_ioctl
};


static int __init bossa_manager_ioctl_init(void) {
  // initialize /dev directory
  devfs_handle = devfs_mk_dir (NULL, "bossa", NULL);
  major = devfs_register_chrdev(0, BOSSA_DEVICE_NAME, &procfops);
  if (major < 0) {
    printk("devfs_register_chrdev failed\n");
    return -ENODEV;
  }
  devfs_register(devfs_handle, BOSSA_DEVICE_NAME, DEVFS_FL_DEFAULT,
		 major, 0,
		 S_IFCHR | S_IRUGO | S_IWUGO,
		 &procfops, NULL);

  // initialize /proc directory
  proc_bossa = proc_mkdir("bossa", NULL);
  make_static_proc();

  return 0;
}

/*********************************************************************
 *                          Proc function                             *
 *********************************************************************/


/* Initialization of static scheduler */
static void make_static_proc(void) {
  int i;
  for (i=0; i!=schedulers; i++) {
    bossa_create_proc_entry(&all_schedulers[i], proc_bossa);
  }
}

static void bossa_create_proc_entry(struct scheduler_info *sched,
				    struct proc_dir_entry *dir) {
  char name[MAX_SCHED_NAME_LENGTH+5];
  strcpy(name, sched->ops->name);
  strcat(name, ".info");
/* remove the proc_entry of an unloaded scheduler that becomes loaded */
  if (sched->father_proc_entry == proc_bossa &&
      sched->sched_proc_entry == NULL)
    bossa_remove_proc_entry(sched,false,false);
  
  if (sched->sched_proc_entry == NULL) {
/* create proc directory if sched is a virtual scheduler */
    if((sched->ops->type == VS && sched->loaded) || sched->root)
      sched->sched_proc_entry = proc_mkdir(sched->ops->name, dir);
    /* create proc file sched.info */
    create_proc_read_entry(name, 00444, dir,
			   sched->ops->list_sched_info, NULL);
    sched->father_proc_entry = dir;
    /* create a symlink sched.info in used proc directory */
    if (dir != proc_bossa && sched->loaded) {
      char path[MAX_SCHED_PATH_LENGTH];
      find_path_symlink(sched, path, name);
      proc_symlink(name, proc_bossa, path);
    }
  }
}

static void find_path_symlink(struct scheduler_info *sched_child,
			      char *path, char *name) {
  int size = 0;
  struct scheduler_operations *child = sched_child->ops;
  struct next_list_struct *sched_gen;

  for (sched_gen = child->next_list.next;
       sched_gen->sched->ops != child;
       sched_gen = sched_gen->next)
    size += sprintf(path+size,"%s/",sched_gen->sched->ops->name);
  sprintf(path+size,"%s",name);
}

static
struct scheduler_info *find_sched_info(struct scheduler_operations *ops) {
  int i;
  for(i=0; i!=schedulers; i++) {
    if (all_schedulers[i].ops == ops)
      return &all_schedulers[i];
  }
  return NULL;
}

/* to "reload" a proc_file hierarchy*/
static void bossa_create_tree_entry(struct scheduler_info *sched_child) {
  struct scheduler_operations *child = sched_child->ops;
  struct next_list_struct *sched_gen;
  struct proc_dir_entry *dir = proc_bossa;
  struct scheduler_info *sched;

  for (sched_gen = child->next_list.next;
       sched_gen->sched->ops != child;
       sched_gen = sched_gen->next) {
    sched = find_sched_info(sched_gen->sched->ops);
    bossa_create_proc_entry(sched, dir);
    dir = sched->sched_proc_entry;
  }
  /* end of hierarchy */
  bossa_create_proc_entry(sched_child, dir);
}

static void bossa_change_proc_root(struct scheduler_info *sched) {
  int i;
  /* delete the previous root from proc_bossa */
  bossa_remove_proc_entry(sched,true,true);
  /* find a leaf (a process scheduler loaded) and build the hierarchical
     tree from it */
  for (i=0; i!=schedulers; i++) {
  /* sched loaded and type == PS*/
    if((all_schedulers[i].loaded) && (all_schedulers[i].ops->type == PS)) {
      // create everything down to the scheduler
      bossa_create_tree_entry(&all_schedulers[i]);
    }
  }
}

/* verify if the proc_entry has a child and if yes delete it*/
static void find_child_entry(struct scheduler_info * sched){
  int i;
  for(i=0;i!=schedulers;i++) {
    if(all_schedulers[i].father_proc_entry == sched->sched_proc_entry)
      bossa_remove_proc_entry(&all_schedulers[i],true,false);
  }
}

static void bossa_remove_proc_entry(struct scheduler_info *sched,
				    int was_loaded, int was_root) {
  char name[MAX_SCHED_NAME_LENGTH+5];
  strcpy(name,sched->ops->name);
  strcat(name, ".info");
  if (sched->sched_proc_entry != NULL)
    find_child_entry(sched);
  /* delete the file .info */
  remove_proc_entry(name, sched->father_proc_entry);
  /* delete the symlink of name */
  remove_proc_entry(name, proc_bossa);
  if((sched->ops->type == VS && was_loaded) || was_root)
    /* delete the sched directory */
    remove_proc_entry(sched->ops->name, sched->father_proc_entry);
  
  sched->father_proc_entry = NULL;
  sched->sched_proc_entry = NULL;
}

/*********************** End of proc function **********************/


/* ---------------------------------------------------------------------
 * exported symbols
 * --------------------------------------------------------------------- */
/* Export symbols so that can be visible to modules */


EXPORT_SYMBOL(bossa_notify_event);

// for modules
EXPORT_SYMBOL(running);

EXPORT_SYMBOL(bossa_add_scheduler);
EXPORT_SYMBOL(bossa_remove_scheduler);
EXPORT_SYMBOL(errno);
EXPORT_SYMBOL(no_next_list);
EXPORT_SYMBOL(rts_idle_process);
EXPORT_SYMBOL(root_s_state);
EXPORT_SYMBOL(bossa_scheduler_lock);

// for modules that use timers
EXPORT_SYMBOL(init_bossa_timer);
EXPORT_SYMBOL(reset_bossa_timer);
EXPORT_SYMBOL(process_bossa_timeout);
EXPORT_SYMBOL(end_bossa_timer);
EXPORT_SYMBOL(no_timer);
EXPORT_SYMBOL(now);

// for modules that use timers
EXPORT_SYMBOL(create_bossa_timer);
EXPORT_SYMBOL(create_static_bossa_timer);
EXPORT_SYMBOL(create_persistent_bossa_timer);
EXPORT_SYMBOL(globalize_bossa_timer);
EXPORT_SYMBOL(delete_bossa_timer);
EXPORT_SYMBOL(delete_static_bossa_timer);
EXPORT_SYMBOL(start_absolute_timer);
EXPORT_SYMBOL(start_relative_timer);
EXPORT_SYMBOL(stop_timer);
#ifdef CONFIG_BOSSA_HRT
EXPORT_SYMBOL(bossa_tstojiffie);
#endif
EXPORT_SYMBOL(find_scheduler_by_id);
EXPORT_SYMBOL(get_user_int);

